import logging
import os
from typing import List, Optional, Union, Dict, Any

import yaml

try:
    from pydantic.v1 import BaseModel, Field, validator
except ImportError:
    from pydantic import BaseModel, Field, validator
from typing_extensions import Literal

from qcportal.utils import seconds_to_hms


def _make_abs_path(path: Optional[str], base_folder: str, default_filename: Optional[str]) -> Optional[str]:
    # No path specified, no default
    if path is None and default_filename is None:
        return None

    # Path isn't specified, but default is given
    if path is None:
        path = default_filename

    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    if os.path.isabs(path):
        return path
    else:
        path = os.path.join(base_folder, path)
        return os.path.abspath(path)


def update_nested_dict(d, u):
    for k, v in u.items():
        if isinstance(v, dict):
            d[k] = update_nested_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class PackageEnvironmentSettings(BaseModel):
    """
    Environments with installed packages that can be used to run calculations

    The compute manager will query these environments to see what packages are installed, and
    direct appropriate calculations to them.
    """

    use_manager_environment: bool = True
    conda: List[str] = Field([], description="List of conda environments to query for installed packages")
    apptainer: List[str] = Field(
        [], description="List of paths to apptainer/singularity files to query for installed packages"
    )


class ExecutorConfig(BaseModel):
    type: str
    queue_tags: List[str]
    worker_init: List[str] = []

    scratch_directory: Optional[str] = None
    bind_address: Optional[str] = None

    cores_per_worker: int
    memory_per_worker: float

    extra_executor_options: Dict[str, Any] = {}

    environments: PackageEnvironmentSettings = PackageEnvironmentSettings()

    class Config(BaseModel.Config):
        case_insensitive = True
        extra = "forbid"


class CustomExecutorConfig(ExecutorConfig):
    type: Literal["custom"] = "custom"
    path: str


class LocalExecutorConfig(ExecutorConfig):
    type: Literal["local"] = "local"
    max_workers: int


class SlurmExecutorConfig(ExecutorConfig):
    type: Literal["slurm"] = "slurm"

    walltime: str
    exclusive: bool = True
    partition: Optional[str] = None
    account: Optional[str] = None

    workers_per_node: int
    max_nodes: int

    scheduler_options: List[str] = []

    @validator("walltime", pre=True)
    def walltime_must_be_str(cls, v):
        if isinstance(v, int):
            return seconds_to_hms(v)
        else:
            return v


class TorqueExecutorConfig(ExecutorConfig):
    type: Literal["torque"] = "torque"

    walltime: str
    account: Optional[str] = None
    queue: Optional[str] = None

    workers_per_node: int
    max_nodes: int

    scheduler_options: List[str] = []

    @validator("walltime", pre=True)
    def walltime_must_be_str(cls, v):
        if isinstance(v, int):
            return seconds_to_hms(v)
        else:
            return v


class LSFExecutorConfig(ExecutorConfig):
    type: Literal["lsf"] = "lsf"

    walltime: str
    project: Optional[str] = None
    queue: Optional[str] = None

    workers_per_node: int
    max_nodes: int

    request_by_nodes: bool = True
    bsub_redirection: bool = True

    scheduler_options: List[str] = []

    @validator("walltime", pre=True)
    def walltime_must_be_str(cls, v):
        if isinstance(v, int):
            return seconds_to_hms(v)
        else:
            return v


AllExecutorTypes = Union[
    CustomExecutorConfig, LocalExecutorConfig, SlurmExecutorConfig, TorqueExecutorConfig, LSFExecutorConfig
]


class FractalServerSettings(BaseModel):
    """
    Settings pertaining to the Fractal Server you wish to pull tasks from and push completed tasks to. Each manager
    supports exactly 1 Fractal Server to be in communication with, and exactly 1 user on that Fractal Server. These
    can be changed, but only once the Manager is shutdown and the settings changed. Multiple Managers however can be
    started in parallel with each other, but must be done as separate calls to the CLI.

    Caution: The password here is written in plain text, so it is up to the owner/writer of the configuration file
    to ensure its security.
    """

    fractal_uri: str = Field(..., description="Full URI to the Fractal Server you want to connect to")
    username: Optional[str] = Field(
        None,
        description="Username to connect to the Fractal Server with. When not provided, a connection is attempted "
        "as a guest user, which in most default Servers will be unable to return results.",
    )
    password: Optional[str] = Field(
        None, description="Password to authenticate to the Fractal Server with (alongside the `username`)"
    )
    verify: Optional[bool] = Field(None, description="Use Server-side generated SSL certification or not.")

    class Config(BaseModel.Config):
        case_insensitive = True
        extra = "forbid"


class FractalComputeConfig(BaseModel):
    base_folder: str = Field(
        ...,
        description="The base folder to use as the default for some options (logs, etc). Default is the location of the config file.",
    )

    cluster: str = Field(
        ...,
        description="Name of this scheduler to present to the Fractal Server. Descriptive names help the server "
        "identify the manager resource and assists with debugging.",
    )
    loglevel: str = "INFO"
    logfile: Optional[str] = Field(
        None,
        description="Full path to save a log file to, including the filename. If not provided, information will still "
        "be reported to terminal, but not saved. When set, logger information is sent to this file.",
    )
    update_frequency: float = Field(
        30,
        description="Time between heartbeats/update checks between this Manager and the Fractal Server. The lower this "
        "value, the shorter the intervals. If you have an unreliable network connection, consider "
        "increasing this time as repeated, consecutive network failures will cause the Manager to shut "
        "itself down to maintain integrity between it and the Fractal Server. Units of seconds",
        gt=0,
    )

    parsl_run_dir: str = "parsl_run_dir"

    server: FractalServerSettings = Field(...)
    environments: PackageEnvironmentSettings = PackageEnvironmentSettings()
    executors: Dict[str, AllExecutorTypes] = Field(...)

    class Config(BaseModel.Config):
        case_insensitive = True
        extra = "forbid"

    @validator("logfile")
    def _check_logfile(cls, v, values):
        return _make_abs_path(v, values["base_folder"], None)

    @validator("parsl_run_dir")
    def _check_run_dir(cls, v, values):
        return _make_abs_path(v, values["base_folder"], "parsl_run_dir")


def read_configuration(file_paths: List[str], extra_config: Optional[Dict[str, Any]] = None) -> FractalComputeConfig:
    logger = logging.getLogger(__name__)
    config_data: Dict[str, Any] = {}

    # Read all the files, in order
    for path in file_paths:
        with open(path, "r") as yf:
            logger.info(f"Reading configuration data from {path}")
            file_data = yaml.safe_load(yf)
            update_nested_dict(config_data, file_data)

    if extra_config:
        update_nested_dict(config_data, extra_config)

    # Find the base folder
    # 1. If specified in the environment, use that
    # 2. Use any specified in a config file
    # 3. Use the path of the last (highest-priority) file given
    if "QCF_COMPUTE_BASE_FOLDER" in os.environ:
        base_dir = os.getenv("QCF_COMPUTE_BASE_FOLDER")
    elif config_data.get("base_folder") is not None:
        base_dir = config_data["base_folder"]
    elif len(file_paths) > 0:
        # use the location of the last file as the base directory
        base_dir = os.path.dirname(file_paths[-1])
    else:
        raise RuntimeError(
            "Base folder must be specified somehow. Maybe set QCF_COMPUTE_BASE_FOLDER in the environment?"
        )

    config_data["base_folder"] = os.path.abspath(base_dir)

    try:
        return FractalComputeConfig(**config_data)
    except Exception as e:
        if len(file_paths) == 0:
            raise RuntimeError(f"Could not assemble a working configuration from environment variables:\n{str(e)}")
        raise



import io
import json
import logging
import re
import sys
import time
import traceback
from contextlib import contextmanager, redirect_stderr, redirect_stdout

import geometric


@contextmanager
def capture_all_output(top_logger: str):
    """Captures all output, including stdout, stderr, and logging"""

    stdout_io = io.StringIO()
    stderr_io = io.StringIO()

    logger = logging.getLogger(top_logger)
    old_handlers = logger.handlers.copy()
    old_prop = logger.propagate

    logger.handlers.clear()
    logger.propagate = False

    # Make logging go to the string io
    handler = logging.StreamHandler(stdout_io)
    handler.terminator = ""
    logger.addHandler(handler)

    # Also redirect stdout/stderr to the string io objects
    with redirect_stdout(stdout_io) as rdout, redirect_stderr(stderr_io) as rderr:
        yield rdout, rderr

        logger.handlers.clear()
        logger.handlers = old_handlers
        logger.propagate = old_prop


if __name__ == "__main__":
    record_id = sys.argv[1]
    nextchain_info_file = sys.argv[2]

    with open(nextchain_info_file, "r") as f:
        nextchain_kwargs = json.load(f)

    with capture_all_output("geometric.nifty") as (rdout, rderr):
        start_time = time.time()
        success = False
        try:
            results = geometric.qcf_neb.nextchain(**nextchain_kwargs)
            success = True
        except Exception as e:
            success = False
            results = {
                "error": {
                    "error_type": type(e).__name__,
                    "error_message": traceback.format_exc(),
                },
                "success": False,
            }

        end_time = time.time()

    stdout = rdout.getvalue()
    stdout = re.sub("Custom engine selected.\n", "", stdout)
    stderr = rderr.getvalue()

    if success:
        ret = {
            "schema_name": "qca_generic_task_result",
            "id": record_id,
            "success": True,
            "stdout": stdout if stdout else None,  # convert empty string to None
            "stderr": stderr if stderr else None,
            "results": results,
            "extras": {},
            "provenance": {
                "creator": "geometric.qcf_neb.nextchain",
                "version": geometric.__version__,
                "walltime": (end_time - start_time),
            },
        }

        print(json.dumps(ret))
    else:
        results["error"]["error_message"] += f"\nstdout: {stdout}\n"
        results["error"]["error_message"] += f"\nstderr: {stderr}\n"
        print(json.dumps(results))



from __future__ import annotations

import logging
import sched
import socket
import threading
import time
import traceback
import uuid
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Tuple

import parsl.executors.high_throughput.interchange
import tabulate
from packaging.version import parse as parse_version
from parsl.config import Config as ParslConfig
from parsl.dataflow.dflow import DataFlowKernel
from parsl.dataflow.futures import Future as ParslFuture
from parsl.executors import HighThroughputExecutor, ThreadPoolExecutor

try:
    from pydantic.v1 import BaseModel, Extra, Field
except ImportError:
    from pydantic import BaseModel, Extra, Field
from requests.exceptions import Timeout

from qcfractalcompute.apps.app_manager import AppManager
from qcportal import ManagerClient
from qcportal.managers import ManagerName
from qcportal.metadata_models import TaskReturnMetadata
from qcportal.record_models import RecordTask
from qcportal.utils import seconds_to_hms
from . import __version__
from .apps.models import AppTaskResult
from .compress import compress_result
from .config import FractalComputeConfig
from .executors import build_executor

if TYPE_CHECKING:
    from parsl.executors.base import ParslExecutor


class InterruptableScheduler(sched.scheduler):
    """
    A scheduler that can be interrupted
    """

    def __init__(self):
        self._int_event = threading.Event()
        super().__init__(time.monotonic, self._int_event.wait)

    def interrupt(self):
        # Clear all events in the queue, then interrupt the current sleep period
        for event in self.queue:
            self.cancel(event)

        self._int_event.set()


class ManagerStatistics(BaseModel):
    """
    Manager statistics
    """

    class Config(BaseModel.Config):
        extra = Extra.forbid

    last_update_time: float = Field(default_factory=time.time)

    active_tasks: int = 0
    active_cores: int = 0
    active_memory: float = 0.0
    total_cpu_hours: float = 0.0

    total_successful_tasks: int = 0
    total_failed_tasks: int = 0
    total_rejected_tasks: int = 0

    @property
    def total_finished_tasks(self) -> int:
        return self.total_successful_tasks + self.total_failed_tasks


class ComputeManager:
    """
    This object maintains a computational queue and watches for finished tasks for different
    queue backends. Finished tasks are added to the database and removed from the queue.
    """

    def __init__(self, config: FractalComputeConfig):
        # Setup logging
        self.logger = logging.getLogger("ComputeManager")

        self.name_data = ManagerName(cluster=config.cluster, hostname=socket.gethostname(), uuid=str(uuid.uuid4()))

        self.client = ManagerClient(
            name_data=self.name_data,
            address=config.server.fractal_uri,
            username=config.server.username,
            password=config.server.password,
            verify=config.server.verify,
        )

        # Check the allowed manager versions
        manager_version_lower_limit = parse_version(self.client.server_info["manager_version_lower_limit"])
        manager_version_upper_limit = parse_version(self.client.server_info["manager_version_upper_limit"])

        manager_version = parse_version(__version__)

        if not manager_version_lower_limit <= manager_version <= manager_version_upper_limit:
            raise RuntimeError(
                f"This manager version {str(manager_version)} does not fall within the server's allowed "
                f"manager versions of [{str(manager_version_lower_limit)}, {str(manager_version_upper_limit)}]."
                f"You may need to upgrade or downgrade"
            )

        # Load the executors
        self.manager_config = config

        self.statistics = ManagerStatistics()

        self.scheduler = InterruptableScheduler()

        # key = number of retries. value = dict of (task_id, compressed result)
        self._deferred_tasks: Dict[int, Dict[int, AppTaskResult]] = defaultdict(dict)

        # key = executor label, value = (key = task_id, value = parsl future)
        self._task_futures: Dict[str, Dict[int, ParslFuture]] = {exl: {} for exl in config.executors.keys()}

        # Mapping of task_id to record_id
        self._record_id_map: Dict[int, int] = {}

        self.all_queue_tags = []
        for ex_config in config.executors.values():
            self.all_queue_tags.extend(ex_config.queue_tags)

        # These are more properly set up in the start() method
        self.parsl_config = None
        self.dflow_kernel = None

        # Set up the app manager
        self.app_manager = AppManager(self.manager_config)
        self.executor_programs = {
            ex: self.app_manager.all_program_info(ex) for ex in self.manager_config.executors.keys()
        }
        self.all_program_info = self.app_manager.all_program_info()

        self.logger.info("-" * 80)
        self.logger.info("QCFractal Compute Manager:")
        self.logger.info("    Version:     {}".format(__version__))
        self.logger.info("    Base folder: {}".format(self.manager_config.base_folder))
        self.logger.info("    Cluster:     {}".format(self.name_data.cluster))
        self.logger.info("    Hostname:    {}".format(self.name_data.hostname))
        self.logger.info("    UUID:        {}".format(self.name_data.uuid))

        self.logger.info("\n")

        self.logger.info("    Parsl Executors:")
        for label, ex_config in config.executors.items():
            self.logger.info("    {}:".format(label))
            self.logger.info("                    Type: {}".format(ex_config.type))
            self.logger.info("              Queue Tags: {}".format(ex_config.queue_tags))
            self.logger.info("                Programs:")
            executor_programs = self.app_manager.all_program_info(label)
            for program, info in executor_programs.items():
                self.logger.info("                    {}: {}".format(program, info))

        self.logger.info("\n")

        # Pull server info
        self.server_info = self.client.get_server_information()
        self.heartbeat_frequency = self.server_info["manager_heartbeat_frequency"]

        self.client.activate(__version__, self.all_program_info, tags=self.all_queue_tags)

        self.logger.info("    Connected to:")
        self.logger.info("        Address:     {}".format(self.client.address))
        self.logger.info("        Name:        {}".format(self.server_info["name"]))
        self.logger.info("        Version:     {}".format(self.server_info["version"]))
        self.logger.info("        Username:    {}".format(self.client.username))
        self.logger.info("        Heartbeat:   {}".format(self.heartbeat_frequency))
        self.logger.info("-" * 80)

        # Is this manager in the process of shutting down?
        # This is used to prevent the manager from starting new tasks in the scheduler
        self._is_stopping = False

        # Number of failed heartbeats. After missing a bunch, we will shutdown
        self._failed_heartbeats = 0

    @staticmethod
    def _get_max_workers(executor: ParslExecutor) -> int:
        """
        Obtain the maximum number of tasks that can be run on an executor
        """

        ####################################################################################
        # This function is broken out slightly, as there are some combinations
        # of executors and providers that are not yet supported (but may be in the future)
        ####################################################################################

        if isinstance(executor, ThreadPoolExecutor):
            return executor.max_threads

        if isinstance(executor, HighThroughputExecutor):
            prov = executor.provider

            # The maximum number of workers are there on a single node
            # (this is somewhat misnamed sometimes. The 'max_workers' also represents
            # the maximum number of workers *per node*, which is set by the user. The
            # 'workers_per_node' is calculated by the executor and takes into account
            # memory and cpu limitations
            workers_per_node = executor.workers_per_node

            # Block information comes from the provider
            # I *think* all providers have a max_blocks and nodes_per_block
            max_blocks = prov.max_blocks
            nodes_per_block = prov.nodes_per_block

            max_tasks = workers_per_node * nodes_per_block * max_blocks
            return max_tasks

        raise ValueError(f"Executor type not supported: {type(executor)}")

    @property
    def name(self) -> str:
        """
        Returns this manager's full name.
        """
        return self.name_data.fullname

    @property
    def n_active_tasks(self) -> Dict[str, int]:
        return {
            ex_label: sum(0 if task.done() else 1 for task in self._task_futures[ex_label].values())
            for ex_label in self._task_futures.keys()
        }

    @property
    def n_total_active_tasks(self) -> int:
        return sum(self.n_active_tasks.values())

    @property
    def n_deferred_tasks(self) -> int:
        return sum(len(x) for x in self._deferred_tasks.values())

    def start(self, manual_updates: bool = False):
        """
        Starts the manager

        This will block until stop() is called
        """

        ###########################################
        # Set up Parsl executors and DataFlowKernel
        ###########################################
        self.parsl_config = ParslConfig(
            executors=[], initialize_logging=False, run_dir=self.manager_config.parsl_run_dir
        )
        self.dflow_kernel = DataFlowKernel(self.parsl_config)

        for ex_label, ex_config in self.manager_config.executors.items():
            ex = build_executor(ex_label, ex_config)
            self.dflow_kernel.add_executors([ex])

        def scheduler_update():
            if not manual_updates:
                self.update(new_tasks=True)
            if not self._is_stopping:
                self.scheduler.enter(self.manager_config.update_frequency, 1, scheduler_update)

        def scheduler_heartbeat():
            if not manual_updates:
                self.heartbeat()
            if not self._is_stopping:
                self.scheduler.enter(self.heartbeat_frequency, 1, scheduler_heartbeat)

        self.logger.info("Compute Manager successfully started.")

        self._failed_heartbeats = 0
        self.scheduler.enter(0, 1, scheduler_update)
        self.scheduler.enter(0, 2, scheduler_heartbeat)

        # Blocks until the ComputeManager.stop() method is called
        self.scheduler.run(blocking=True)

        #############################################
        # If we got here, the scheduler has stopped
        # Now handle the shutdown
        #############################################
        self.update(new_tasks=False)

        try:
            # Notify the server of shutdown
            self.client.deactivate(
                active_tasks=self.statistics.active_tasks,
                active_cores=self.statistics.active_cores,
                active_memory=self.statistics.active_memory,
                total_cpu_hours=self.statistics.total_cpu_hours,
            )

            shutdown_str = "Shutdown was successful, {} tasks returned to the fractal server"

        except Exception as ex:
            self.logger.warning(f"Deactivation failed for {self.name}: {str(ex).strip()}")
            shutdown_str = "Shutdown was not successful, {} tasks not returned."

        shutdown_str = shutdown_str.format(f"{self.n_total_active_tasks} active and {self.n_deferred_tasks} stale")
        self.logger.info(f"Manager {self.name}: {shutdown_str}")

        # Close down the parsl stuff. Can sometimes get called after the
        # DataFlowKernel atexit handler has been called
        if not self.dflow_kernel.cleanup_called:
            self.dflow_kernel.cleanup()

        self.dflow_kernel = None
        self.parsl_config = None

        self.logger.info("Compute manager stopping gracefully.")

    def stop(self) -> None:
        """
        Interrupts a running worker, causing it to shut down
        """
        self.logger.info("Manager stopping")

        self._is_stopping = True

        # This interrupts the scheduler, which will cause the rest of the start() method to run
        self.scheduler.interrupt()

    def heartbeat(self) -> None:
        """
        Provides a heartbeat to the connected Server.
        """

        try:
            self.client.heartbeat(
                active_tasks=self.statistics.active_tasks,
                active_cores=self.statistics.active_cores,
                active_memory=self.statistics.active_memory,
                total_cpu_hours=self.statistics.total_cpu_hours,
            )
            self._failed_heartbeats = 0

        except (ConnectionError, Timeout) as ex:
            self._failed_heartbeats += 1
            self.logger.warning(f"Heartbeat failed: {str(ex).strip()}. QCFractal server down?")
            self.logger.warning(f"Missed {self._failed_heartbeats} heartbeats so far")
            if self._failed_heartbeats > self.client.server_info["manager_heartbeat_max_missed"]:
                self.logger.warning("Too many failed heartbeats, shutting down.")
                self.stop()

    def _acquire_complete_tasks(self) -> Dict[str, Dict[int, AppTaskResult]]:
        # First key is name of executor
        # Second key is task_id
        # Value is the result (including compressed computation result)
        ret: Dict[str, Dict[int, AppTaskResult]] = {}

        for executor_label, task_futures in self._task_futures.items():
            ret.setdefault(executor_label, {})

            # Finished tasks - will be removed from the task_futures dict later
            finished: List[int] = []

            for task_id, task in task_futures.items():
                if task.done():
                    try:
                        ret[executor_label][task_id] = task.result()

                    except parsl.executors.high_throughput.interchange.ManagerLost as e:
                        msg = "Compute worker lost:\n" + traceback.format_exc()
                        failed_op = {
                            "success": False,
                            "error": {"error_type": e.__class__.__name__, "error_message": msg},
                        }

                        ret[executor_label][task_id] = AppTaskResult(
                            success=False, walltime=0.0, result_compressed=compress_result(failed_op)
                        )

                    except Exception as e:
                        msg = "Error getting task result:\n" + traceback.format_exc()
                        failed_op = {
                            "success": False,
                            "error": {"error_type": e.__class__.__name__, "error_message": msg},
                        }

                        ret[executor_label][task_id] = AppTaskResult(
                            success=False, walltime=0.0, result_compressed=compress_result(failed_op)
                        )

                    finished.append(task_id)

            for task_id in finished:
                del task_futures[task_id]

        # Print out the table of finished tasks
        # columns: task_id, record_id, executor, walltime, status
        table_rows: List[Tuple[int, int, str, str, str]] = []
        for executor_label, executor_results in ret.items():
            for task_id, task_result in executor_results.items():
                if task_result.success:
                    status_str = "success"
                else:
                    status_str = "error: " + task_result.result["error"]["error_type"]

                table_rows.append(
                    (
                        task_id,
                        self._record_id_map[task_id],
                        executor_label,
                        seconds_to_hms(task_result.walltime),
                        status_str,
                    )
                )

        log_str = f"Acquired {len(table_rows)} finished tasks from the executors"

        if table_rows:
            log_str += "\n" + tabulate.tabulate(
                sorted(table_rows), headers=["task id", "record id", "executor", "walltime", "status"]
            )

        self.logger.info(log_str)

        return ret

    def _submit_tasks(self, executor_label: str, tasks: List[RecordTask]):
        """
        Submits tasks to the parsl queue to be run
        """

        for task in tasks:
            task_app = self.app_manager.get_app(self.dflow_kernel, executor_label, task)
            task_future = task_app(
                task.record_id,
                task.function_kwargs_compressed,
                executor_config=self.manager_config.executors[executor_label],
            )
            self._task_futures[executor_label][task.id] = task_future
            self._record_id_map[task.id] = task.record_id

    def _return_finished(self, results: Dict[int, AppTaskResult]) -> TaskReturnMetadata:
        # Handling of exceptions is expected to be done in the calling function
        to_send = {k: v.result_compressed for k, v in results.items()}
        return_meta = self.client.return_finished(to_send)

        if return_meta.success:
            self.logger.info(f"Successfully returned {return_meta.n_accepted} tasks to the fractal server")
        else:
            self.logger.warning(f"Error in returning tasks: {str(return_meta.error_string)}")

        return return_meta

    def _update_deferred_tasks(self) -> Dict[int, TaskReturnMetadata]:
        """
        Attempt to post the previous payload failures
        """
        new_deferred_tasks = defaultdict(dict)

        # key = number of attempts. value = metadata
        ret: Dict[int, TaskReturnMetadata] = {}

        for attempts, results in self._deferred_tasks.items():
            try:
                return_meta = self._return_finished(results)
                ret[attempts] = return_meta

                if return_meta.success:
                    self.logger.info(f"Successfully pushed jobs from {attempts+1} updates ago")
                else:
                    self.logger.warning(
                        f"Did not successfully push jobs from {attempts+1} updates ago. Error: {return_meta.error_string}"
                    )

            except (Timeout, ConnectionError):
                # Tried and failed
                attempts += 1

                new_deferred_tasks[attempts] = results
                self.logger.warning(f"Could not post jobs from {attempts-1} updates ago, will retry on next update.")

        self._deferred_tasks = new_deferred_tasks
        return ret

    def update(self, new_tasks) -> None:
        """Examines the queue for completed tasks and adds successful completions to the database
        while unsuccessful are logged for future inspection.

        Parameters
        ----------
        new_tasks
            Try to get new tasks from the server
        """

        # First, try pushing back any stale results
        deferred_return_info = self._update_deferred_tasks()

        results = self._acquire_complete_tasks()

        server_up = True

        # Stores rows of the status table printed at the end
        # Columns: task_id, status, reason
        # record_id will be added later
        status_rows: List[Tuple[int, str, str]] = []

        # Add the info from updating deferred tasks to the table
        for attempts, return_meta in deferred_return_info.items():
            status_rows.extend(
                [(task_id, f"sent (was deferred {attempts})", "") for task_id in return_meta.accepted_ids]
            )
            status_rows.extend(
                [
                    (task_id, f"rejected (was deferred {attempts})", reason)
                    for task_id, reason in return_meta.rejected_info
                ]
            )
            self.statistics.total_rejected_tasks += return_meta.n_rejected

        # Return results to the server (per executor)
        for executor_label, executor_results in results.items():
            # Any post-processing tasks
            # Sometimes used for saving data for later
            self.postprocess_results(executor_results)

            n_result = len(executor_results)

            if n_result:
                n_success = 0

                try:
                    return_meta = self._return_finished(executor_results)

                    status_rows.extend([(task_id, "sent", "") for task_id in return_meta.accepted_ids])

                    status_rows.extend([(task_id, "rejected", reason) for task_id, reason in return_meta.rejected_info])
                    self.statistics.total_rejected_tasks += return_meta.n_rejected

                except (ConnectionError, Timeout):
                    self.logger.warning("Returning complete tasks failed. Attempting again on next update.")
                    self._deferred_tasks[0].update(executor_results)

                    status_rows.extend([(task_id, "deferred", "") for task_id in executor_results.keys()])
                    server_up = False

                for task_id, app_result in executor_results.items():
                    walltime_seconds = app_result.walltime

                    if app_result.success:
                        n_success += 1
                    else:
                        self.logger.debug(f"Task {task_id} (record {self._record_id_map[task_id]}) failed:")
                        self.logger.debug(app_result.result["error"]["error_message"])

                    cores_per_worker = self.manager_config.executors[executor_label].cores_per_worker
                    self.statistics.total_cpu_hours += walltime_seconds * cores_per_worker / 3600

                n_fail = n_result - n_success

                self.logger.info(
                    f"Executor {executor_label}: Processed {n_result} tasks: {n_success} success / {n_fail} failed"
                )

                # Update the statistics
                self.statistics.total_successful_tasks += n_success
                self.statistics.total_failed_tasks += n_fail

        ########################################################################
        # Update a few more statistics
        # total_successful_tasks/n_failed_tasks are updated above, per executor
        ########################################################################

        n_active_tasks = self.n_active_tasks

        # Active cores - active tasks * cores per task (worker) for each executor
        self.statistics.active_cores = sum(
            n_active_tasks[ex_label] * ex_config.cores_per_worker
            for ex_label, ex_config in self.manager_config.executors.items()
        )

        # Same as above, but for memory
        self.statistics.active_memory = sum(
            n_active_tasks[ex_label] * ex_config.memory_per_worker
            for ex_label, ex_config in self.manager_config.executors.items()
        )

        if status_rows:
            log_str = "Task return status:\n"

            # Add record_id
            new_status_rows = [
                (task_id, self._record_id_map[task_id], status, reason)
                for task_id, status, reason in sorted(status_rows)
            ]
            log_str += tabulate.tabulate(new_status_rows, headers=["task id", "record id", "status", "reason"])
            self.logger.info(log_str)

        ###########################################
        # Write statistics to the log
        #######################################
        task_stats_str = (
            f"Task Stats: Total finished={self.statistics.total_finished_tasks}, "
            f"Failed={self.statistics.total_failed_tasks}, "
            f"Success={self.statistics.total_successful_tasks}, "
            f"Rejected={self.statistics.total_rejected_tasks}"
        )

        worker_stats_str = f"Worker Stats (est.): Core Hours Used={self.statistics.total_cpu_hours:,.2f}"

        self.logger.info(task_stats_str)
        self.logger.info(worker_stats_str)
        self.statistics.last_update_time = time.time()

        if new_tasks and server_up:
            # What do we have for each executor?
            active_tasks = self.n_active_tasks

            for executor_label, executor_config in self.manager_config.executors.items():
                executor = self.dflow_kernel.executors[executor_label]

                # How many slots do we have?
                # TODO - intelligently figure out the number tasks to claim over the number of slots
                open_slots = (3 * self._get_max_workers(executor)) - active_tasks[executor_label]

                self.logger.info(
                    f"Executor {executor_label} has {active_tasks[executor_label]} active tasks and {open_slots} open slots"
                )

                if open_slots > 0:
                    try:
                        executor_programs = self.executor_programs[executor_label]
                        new_task_info = self.client.claim(executor_programs, executor_config.queue_tags, open_slots)
                    except (Timeout, ConnectionError) as ex:
                        self.logger.warning(f"Acquisition of new tasks failed: {str(ex).strip()}")
                        return

                    self.logger.info("Acquired {} new tasks.".format(len(new_task_info)))

                    # Add new tasks to queue
                    self.preprocess_new_tasks(new_task_info)
                    self._submit_tasks(executor_label, new_task_info)

    def preprocess_new_tasks(self, new_tasks: List[RecordTask]):
        """
        Any processing to do to the new tasks

        To be overridden by a derived class. Sometimes used to save the results for testing
        """
        pass

    def postprocess_results(self, results: Dict[int, AppTaskResult]):
        """
        Any processing to do to the results

        To be overridden by a derived class. Sometimes used to save the results for testing
        """
        pass



from __future__ import annotations

import threading
import time
from typing import TYPE_CHECKING

import pytest

from qcfractalcompute.compute_manager import ComputeManager
from qcfractalcompute.config import FractalComputeConfig, FractalServerSettings, LocalExecutorConfig
from qcfractalcompute.testing_helpers import QCATestingComputeThread, populate_db
from qcportal.managers import ManagerStatusEnum, ManagerQueryFilters
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


def test_manager_keepalive(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()

    snowflake.start_job_runner()

    compute = QCATestingComputeThread(snowflake._qcf_config, {})
    compute.start(manual_updates=True)

    time.sleep(1)  # wait for manager to register

    managers = storage_socket.managers.query(ManagerQueryFilters())
    assert len(managers) == 1
    manager_name = managers[0]["name"]

    sleep_time = snowflake._qcf_config.heartbeat_frequency
    max_missed = snowflake._qcf_config.heartbeat_max_missed

    for i in range(max_missed * 2):
        time_0 = now_at_utc()
        compute._compute.heartbeat()
        time_1 = now_at_utc()
        time.sleep(sleep_time)
        m = storage_socket.managers.get([manager_name])
        assert m[0]["status"] == ManagerStatusEnum.active
        assert time_0 < m[0]["modified_on"] < time_1

    # No more updates, server should eventually mark as inactive
    time.sleep(sleep_time * (max_missed + 1))
    m = storage_socket.managers.get([manager_name])
    assert m[0]["status"] == ManagerStatusEnum.inactive


def test_manager_tags(snowflake: QCATestingSnowflake, tmp_path):
    storage_socket = snowflake.get_storage_socket()

    compute_config = FractalComputeConfig(
        base_folder=str(tmp_path),
        cluster="testing_compute",
        update_frequency=5,
        server=FractalServerSettings(
            fractal_uri=snowflake.get_uri(),
            verify=False,
        ),
        executors={
            "local": LocalExecutorConfig(
                cores_per_worker=1,
                memory_per_worker=1,
                max_workers=1,
                queue_tags=["tag1", "tag2", "*"],
            ),
            "local2": LocalExecutorConfig(
                cores_per_worker=1, memory_per_worker=1, max_workers=1, queue_tags=["tag3", "tag4"]
            ),
        },
    )

    compute = ComputeManager(compute_config)
    compute_thread = threading.Thread(target=compute.start)
    compute_thread.start()
    time.sleep(2)
    compute.stop()
    compute_thread.join()

    managers = storage_socket.managers.query(ManagerQueryFilters())
    assert len(managers) == 1
    assert set(managers[0]["tags"]) == {"tag1", "tag2", "tag3", "tag4", "*"}


@pytest.mark.filterwarnings("ignore:Exception in thread")
def test_manager_claim_inactive(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake.start_job_runner()

    compute = QCATestingComputeThread(snowflake._qcf_config, {})
    compute.start(manual_updates=False)

    time.sleep(2)  # wait for manager to register
    assert compute.is_alive() is True

    managers = storage_socket.managers.query(ManagerQueryFilters())
    assert len(managers) == 1
    manager_name = managers[0]["name"]

    # Mark as inactive from the server side
    storage_socket.managers.deactivate([manager_name])

    # Next update should kill the process
    time.sleep(compute._compute._compute_config.update_frequency + 2)

    # Should have killed the manager process
    assert compute.is_alive() is False


def test_manager_claim_return(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    all_id, result_data = populate_db(storage_socket)

    compute = QCATestingComputeThread(snowflake._qcf_config, result_data)
    compute.start(manual_updates=False)

    time.sleep(1)  # wait for manager to register
    assert compute.is_alive() is True

    managers = storage_socket.managers.query(ManagerQueryFilters())
    assert len(managers) == 1

    r = snowflake.await_results(all_id, 30.0)
    assert r is True


def test_manager_deferred_return(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    all_id, result_data = populate_db(storage_socket)

    compute_thread = QCATestingComputeThread(snowflake._qcf_config, result_data)
    compute_thread.start(manual_updates=True)
    compute = compute_thread._compute

    time.sleep(1)  # wait for manager to register
    managers = storage_socket.managers.query(ManagerQueryFilters())
    assert len(managers) == 1
    assert compute.n_total_active_tasks == 0  # haven't updated - we are doing manual updates

    # Get some tasks
    compute.update(new_tasks=True)
    assert compute.n_total_active_tasks > 0
    assert compute.n_deferred_tasks == 0

    # Sever goes down
    snowflake.stop_api()

    # Let manager complete some tasks
    time.sleep(3)  # Mock testing adapter waits for two seconds before returning result
    compute.update(new_tasks=True)
    assert compute.n_deferred_tasks > 0
    deferred_task_ids = list(compute._deferred_tasks[0].keys())
    deferred_record_ids = [compute._record_id_map[x] for x in deferred_task_ids]

    # Now server comes back
    snowflake.start_api()

    # Manager can now update
    compute.update(new_tasks=True)

    # No more deferred tasks
    assert compute.n_deferred_tasks == 0
    assert compute.n_total_active_tasks > 0

    # Record is complete on the server
    r = storage_socket.records.get(deferred_record_ids)
    assert all(x["status"] == "complete" for x in r)
    assert all(x["manager_name"] == compute.name for x in r)


def test_manager_missed_heartbeats_shutdown(snowflake: QCATestingSnowflake):
    compute_thread = QCATestingComputeThread(snowflake._qcf_config)
    compute_thread.start(manual_updates=False)

    snowflake.stop_api()

    for i in range(60):
        time.sleep(1)

        if not compute_thread.is_alive():
            break
    else:
        raise RuntimeError("Compute thread did not stop in 60 seconds")

    compute_thread._compute_thread.join(5)
    assert compute_thread.is_alive() is False



import json

# This script returns the programs/procedures available in qcengine
# as a dictionary of {program: version}
# It is meant to be used with subprocess to get the available programs
# in a conda environment

if __name__ == "__main__":
    try:
        import qcengine

        progs = {x: qcengine.get_program(x).get_version() for x in qcengine.list_available_programs()}
        procs = {x: qcengine.get_procedure(x).get_version() for x in qcengine.list_available_procedures()}
        progs["qcengine"] = qcengine.__version__

    except ImportError:
        progs = {}
        procs = {}

    print(json.dumps({**progs, **procs}))



from __future__ import annotations

import math
import sys
from typing import TYPE_CHECKING

from parsl.channels import LocalChannel
from parsl.executors import ThreadPoolExecutor, HighThroughputExecutor
from parsl.providers import SlurmProvider, TorqueProvider, LSFProvider

from qcfractalcompute.config import (
    ExecutorConfig,
    LocalExecutorConfig,
    SlurmExecutorConfig,
    TorqueExecutorConfig,
    LSFExecutorConfig,
    CustomExecutorConfig,
)

if TYPE_CHECKING:
    from parsl.executors.base import ParslExecutor


def build_executor(executor_label: str, executor_config: ExecutorConfig) -> ParslExecutor:
    """
    Build a Parsl executor from a given ExecutorConfig
    """

    ##############################################################################
    # General notes
    # -------------
    # Overall, what we call "task" is what Parsl calls a "worker". In Parsl,
    # a worker is responsible for running a single task at a time. So we tend
    # to think of stuff in terms of concurrent tasks rather than workers.
    ##############################################################################

    # Always use local channel
    lchannel = LocalChannel()

    if executor_config.type == "local":
        assert isinstance(executor_config, LocalExecutorConfig)

        # Use a thread pool with a local provider
        # Always use just a single block in the provider
        # Pretty straightforward mapping from config to parsl config
        return ThreadPoolExecutor(
            label=executor_label,
            max_threads=executor_config.max_workers,
            **executor_config.extra_executor_options,
        )

        ####################################################################################
        # The code below may make sense at some point, but closing a HighThroughputExecutor
        # doesn't seem to work properly (it leaves threads/processes running).
        ####################################################################################
        ## Use a HighThroughputExecutor with a local provider
        ## Always use just a single block in the provider
        ## Pretty straightforward mapping from config to parsl config
        # return HighThroughputExecutor(
        #    label=executor_label,
        #    cores_per_worker=executor_config.cores_per_worker,
        #    mem_per_worker=executor_config.memory_per_worker,
        #    max_workers=executor_config.max_workers,
        #    address="127.0.0.1",
        #    provider=LocalProvider(
        #        channel=lchannel, init_blocks=1, max_blocks=1, worker_init=";".join(executor_config.worker_init)
        #    ),
        # )

    elif executor_config.type == "slurm":
        assert isinstance(executor_config, SlurmExecutorConfig)
        # Use a HighThroughputExecutor with a Slurm provider
        # Use blocks of size 1, so number of nodes = number of blocks
        # Pretty straightforward mapping from config to parsl config
        # We let the high-throughput executor handle max_workers (which is the
        # max *per node*, not overall)

        # User specifies resources per worker, so convert to resources per node
        mem_per_node = executor_config.memory_per_worker * executor_config.workers_per_node
        cores_per_node = executor_config.cores_per_worker * executor_config.workers_per_node

        return HighThroughputExecutor(
            label=executor_label,
            cores_per_worker=executor_config.cores_per_worker,
            mem_per_worker=executor_config.memory_per_worker,
            address=executor_config.bind_address,
            provider=SlurmProvider(
                channel=lchannel,
                init_blocks=1,
                min_blocks=0,
                max_blocks=executor_config.max_nodes,
                nodes_per_block=1,
                mem_per_node=math.ceil(mem_per_node),
                cores_per_node=cores_per_node,
                exclusive=executor_config.exclusive,
                walltime=executor_config.walltime,
                account=executor_config.account,
                partition=executor_config.partition,
                worker_init=";".join(executor_config.worker_init),
                scheduler_options="\n".join(executor_config.scheduler_options),
            ),
        )

    elif executor_config.type == "torque":
        assert isinstance(executor_config, TorqueExecutorConfig)
        # Use a HighThroughputExecutor with a Torque provider
        # Use blocks of size 1, so number of nodes = number of blocks
        # Pretty straightforward mapping from config to parsl config
        # We let the high-throughput executor handle max_workers (which is the
        # max *per node*, not overall)

        # User specifies resources per worker, so convert to resources per node
        cores_per_node = executor_config.cores_per_worker * executor_config.workers_per_node

        return HighThroughputExecutor(
            label=executor_label,
            cores_per_worker=executor_config.cores_per_worker,
            mem_per_worker=executor_config.memory_per_worker,
            address=executor_config.bind_address,
            provider=TorqueProvider(
                channel=lchannel,
                init_blocks=1,
                min_blocks=0,
                max_blocks=executor_config.max_nodes,
                nodes_per_block=1,
                walltime=executor_config.walltime,
                account=executor_config.account,
                queue=executor_config.queue,
                worker_init=";".join(executor_config.worker_init),
                scheduler_options="\n".join(executor_config.scheduler_options),
            ),
        )

    elif executor_config.type == "lsf":
        assert isinstance(executor_config, LSFExecutorConfig)
        # Use a HighThroughputExecutor with a LSF provider
        # Use blocks of size 1, so number of nodes = number of blocks
        # Pretty straightforward mapping from config to parsl config
        # We let the high-throughput executor handle max_workers (which is the
        # max *per node*, not overall)

        # User specifies resources per worker, so convert to resources per node
        cores_per_node = executor_config.cores_per_worker * executor_config.workers_per_node

        cores_per_block = None

        if executor_config.request_by_nodes is False:
            # For us, 1 block = 1 node
            cores_per_block = cores_per_node

        return HighThroughputExecutor(
            label=executor_label,
            cores_per_worker=executor_config.cores_per_worker,
            mem_per_worker=executor_config.memory_per_worker,
            address=executor_config.bind_address,
            provider=LSFProvider(
                channel=lchannel,
                init_blocks=1,
                min_blocks=0,
                max_blocks=executor_config.max_nodes,
                cores_per_block=cores_per_block,
                nodes_per_block=1,
                cores_per_node=cores_per_node,
                walltime=executor_config.walltime,
                project=executor_config.project,
                queue=executor_config.queue,
                request_by_nodes=executor_config.request_by_nodes,
                bsub_redirection=executor_config.bsub_redirection,
                worker_init=";".join(executor_config.worker_init),
                scheduler_options="\n".join(executor_config.scheduler_options),
            ),
        )

    elif executor_config.type == "custom":
        assert isinstance(executor_config, CustomExecutorConfig)
        raise RuntimeError("TODO")
        sys.path.append("/home/ben/work/qcarchive/servers/next/new_worker")

        m = importlib.import_module(executor_config.config_file)
        ex = m.get_executor()

        for tag in executor_config.queue_tags:
            self.tag_executor_map[tag] = ex.label

        self.executor_config_map[ex.label] = executor_config
        pass
    else:
        raise ValueError("Unknown executor type: {}".format(executor_config.type))



from __future__ import annotations

import json
import subprocess
import tempfile
from functools import lru_cache, partial
from typing import TYPE_CHECKING, Dict, List, Set, Any, Optional

import parsl

from qcfractalcompute.apps.geometric import geometric_nextchain_conda_app, geometric_nextchain_apptainer_app
from qcfractalcompute.apps.helpers import run_apptainer
from qcfractalcompute.apps.qcengine import qcengine_conda_app, qcengine_apptainer_app
from qcfractalcompute.run_scripts import get_script_path
from qcportal.record_models import RecordTask

if TYPE_CHECKING:
    from parsl.dataflow.dflow import DataFlowKernel
    from qcfractalcompute.config import FractalComputeConfig


@lru_cache()
def discover_programs_conda(conda_env_name: Optional[str]) -> Dict[str, Dict[str, Any]]:
    qcengine_list_path = get_script_path("qcengine_list.py")

    if conda_env_name:
        cmd = ["conda", "run", "-n", conda_env_name, "python3", qcengine_list_path]
    else:
        cmd = ["python3", qcengine_list_path]

    # Use a temporary dir. QCEngine will sometimes write files there (like timer.dat)
    with tempfile.TemporaryDirectory() as tmpdir:
        result = subprocess.check_output(cmd, universal_newlines=True, cwd=tmpdir)

    # QCEngine differentiates between programs and procedures, but we don't
    program_info = json.loads(result)

    # functions are always the same
    functions = {
        "qcengine.compute": {
            "programs": program_info,
            "app_function": partial(qcengine_conda_app, conda_env_name=conda_env_name),
        },
        "qcengine.compute_procedure": {
            "programs": program_info,
            "app_function": partial(qcengine_conda_app, conda_env_name=conda_env_name),
        },
    }

    if "geometric" in program_info:
        functions["geometric.qcf_neb.nextchain"] = {
            "programs": {"geometric": program_info["geometric"]},
            "app_function": partial(geometric_nextchain_conda_app, conda_env_name=conda_env_name),
        }

    return functions


@lru_cache()
def discover_programs_apptainer(sif_path: str) -> Dict[str, Dict[str, Any]]:
    qcengine_list_path = get_script_path("qcengine_list.py")

    program_info = run_apptainer(
        sif_path, command=["python3", "/qcengine_list.py"], volumes=[(qcengine_list_path, "/qcengine_list.py")]
    )

    # functions are always the same
    functions = {
        "qcengine.compute": {
            "programs": program_info,
            "app_function": partial(qcengine_apptainer_app, sif_path=sif_path),
        },
        "qcengine.compute_procedure": {
            "programs": program_info,
            "app_function": partial(qcengine_apptainer_app, sif_path=sif_path),
        },
    }

    if "geometric" in program_info:
        functions["geometric.qcf_neb.nextchain"] = {
            "programs": {"geometric": program_info["geometric"]},
            "app_function": partial(geometric_nextchain_apptainer_app, sif_path=sif_path),
        }

    return functions


class AppManager:
    def __init__(self, manager_config: FractalComputeConfig):
        # key is executor label
        self._parsl_apps = {}

        for executor_label, executor_config in manager_config.executors.items():
            self._parsl_apps[executor_label] = []

            # Merge in the global config into the executor-specific config
            use_current_env = executor_config.environments.use_manager_environment
            conda_envs = set(manager_config.environments.conda) | set(executor_config.environments.conda)

            # Same for apptainers
            apptainers = set(manager_config.environments.apptainer) | set(executor_config.environments.apptainer)

            # Check the current environment
            if use_current_env:
                qcengine_functions = discover_programs_conda(None)
                for qcengine_function_name, func_info in qcengine_functions.items():
                    self._parsl_apps[executor_label].append(
                        (qcengine_function_name, func_info["programs"], func_info["app_function"])
                    )

            # Check the conda environments in the config
            for conda_env in conda_envs:
                qcengine_functions = discover_programs_conda(conda_env)
                for qcengine_function_name, func_info in qcengine_functions.items():
                    self._parsl_apps[executor_label].append(
                        (qcengine_function_name, func_info["programs"], func_info["app_function"])
                    )

            for apptainer in apptainers:
                qcengine_functions = discover_programs_apptainer(apptainer)
                for qcengine_function_name, func_info in qcengine_functions.items():
                    self._parsl_apps[executor_label].append(
                        (qcengine_function_name, func_info["programs"], func_info["app_function"])
                    )

    def get_app(self, dflow_kernel: DataFlowKernel, executor_label: str, task: RecordTask) -> Any:
        task_programs = set(task.required_programs)

        if executor_label not in self._parsl_apps:
            raise KeyError(f"No executor available with label {executor_label}")

        for name, programs, func in self._parsl_apps[executor_label]:
            if task.function == name and task_programs.issubset(programs.keys()):
                return parsl.python_app(func, data_flow_kernel=dflow_kernel, executors=[executor_label])

        raise KeyError(
            f"No app available for executor {executor_label}, function {task.function}, programs {task_programs}"
        )

    def all_program_info(self, executor_label: Optional[str] = None) -> Dict[str, List[str]]:
        """Returns a dictionary of all program information.

        The dictionary is has keys of program name, and values of a list of available versions.
        """

        ret: Dict[str, Set[str]] = {}
        if executor_label is not None:
            for _, program_info, _ in self._parsl_apps[executor_label]:
                for prog, ver in program_info.items():
                    ret.setdefault(prog, set())
                    if ver is None:
                        ver = "unknown"
                    ret[prog].add(ver)
        else:
            for executor_label, executor_info in self._parsl_apps.items():
                for _, program_info, _ in executor_info:
                    for prog, ver in program_info.items():
                        ret.setdefault(prog, set())
                        if ver is None:
                            ver = "unknown"
                        ret[prog].add(ver)

        # Convert versions to a list
        return {k: list(v) for k, v in ret.items()}



import importlib
import operator
import os
import time
from typing import Dict, Any, Union

from qcelemental.models import Provenance, FailedOperation, ComputeError

from qcfractalcompute import __version__
from qcportal.generic_result import GenericTaskResult
from qcportal.utils import capture_all_output

_this_dir = os.path.abspath(os.path.dirname(__file__))
_script_path = os.path.join(_this_dir, "run_scripts/generic_script.py")


def wrap_generic_function(
    record_id: int, function: str, function_kwargs: Dict[str, Any]
) -> Union[GenericTaskResult, FailedOperation]:
    """
    Wraps a generic function call to return a GenericTaskResult (or FailedOperation)

    This function handles capturing stdout/sterr and timing as well. This function takes
    the origin task_info (basically TaskQueueORM as a dictionary) and runs the specified function
    """

    module_name, func_name = function.split(".", 1)
    module = importlib.import_module(module_name)
    func = operator.attrgetter(func_name)(module)

    with capture_all_output("") as (rdout, rderr):
        start_time = time.time()
        try:
            results = func(**function_kwargs)
        except Exception as e:
            err = ComputeError(
                error_type=type(e).__name__,
                error_message=str(e),
            )
            return FailedOperation(id=record_id, error=err)

        end_time = time.time()

    stdout = rdout.getvalue()
    stderr = rderr.getvalue()

    ret = GenericTaskResult(
        id=record_id,
        success=True,
        stdout=stdout if stdout else None,  # convert empty string to None
        stderr=stderr if stderr else None,
        results=results,
        provenance=Provenance(
            creator=__name__,
            version=__version__,
            walltime=(end_time - start_time),
        ),
    )

    return ret



from __future__ import annotations

import os
import tempfile
import threading
import weakref
from typing import Dict, List, Optional

import parsl

from qcarchivetesting.testing_classes import _activated_manager_programs
from qcfractal.components.optimization.testing_helpers import submit_test_data as submit_opt_test_data
from qcfractal.components.singlepoint.testing_helpers import submit_test_data as submit_sp_test_data
from qcfractal.config import FractalConfig
from qcfractal.db_socket import SQLAlchemySocket
from qcfractalcompute.apps.models import AppTaskResult
from qcfractalcompute.compress import compress_result
from qcfractalcompute.compute_manager import ComputeManager
from qcfractalcompute.config import FractalComputeConfig, FractalServerSettings, LocalExecutorConfig
from qcportal.all_results import AllResultTypes
from qcportal.record_models import PriorityEnum, RecordTask


class MockTestingComputeManager(ComputeManager):
    def __init__(self, qcf_config: FractalConfig, result_data: Dict[int, AllResultTypes]):
        self._qcf_config = qcf_config

        host = self._qcf_config.api.host
        port = self._qcf_config.api.port
        uri = f"http://{host}:{port}"

        # Not we set the update_frequency to be quite a bit less than the heartbeat_frequency
        # from the server. This allows us to test dropping deferred tasks without having the manager
        # die due to missed heartbeats
        tmpdir = tempfile.TemporaryDirectory()
        parsl_run_dir = os.path.join(tmpdir.name, "parsl_run_dir")
        compute_scratch_dir = os.path.join(tmpdir.name, "compute_scratch_dir")
        os.makedirs(parsl_run_dir, exist_ok=True)
        os.makedirs(compute_scratch_dir, exist_ok=True)

        self._compute_config = FractalComputeConfig(
            base_folder=tmpdir.name,
            parsl_run_dir=parsl_run_dir,
            cluster="mock_compute",
            update_frequency=1,
            server=FractalServerSettings(
                fractal_uri=uri,
                verify=False,
            ),
            executors={
                "local": LocalExecutorConfig(
                    cores_per_worker=1,
                    memory_per_worker=1,
                    max_workers=1,
                    queue_tags=["*"],
                    scratch_directory=compute_scratch_dir,
                )
            },
        )
        ComputeManager.__init__(self, self._compute_config)

        # overwrite the executor programs for testing
        self.executor_programs = {"local": _activated_manager_programs}

        def cleanup(d):
            d.cleanup()

        weakref.finalize(self, cleanup, tmpdir)

        # Shorten the timeout on the client for testing
        self.client._timeout = 2

        self._result_data = result_data

    # We have an executor and everything running, but we short-circuit the actual compute
    def _submit_tasks(self, executor_label: str, tasks: List[RecordTask]):
        # A mock app that just returns the result data given to it after two seconds
        @parsl.python_app(data_flow_kernel=self.dflow_kernel)
        def _mock_app(result: AllResultTypes) -> AppTaskResult:
            import time

            time.sleep(2)
            return AppTaskResult(success=result.success, walltime=2.0, result_compressed=compress_result(result.dict()))

        for task in tasks:
            self._record_id_map[task.id] = task.record_id
            task_future = _mock_app(self._result_data[task.record_id])
            self._task_futures[executor_label][task.id] = task_future


class QCATestingComputeThread:
    """
    Runs a compute manager in a separate process
    """

    def __init__(self, qcf_config: FractalConfig, result_data: Dict[int, AllResultTypes] = None):
        self._qcf_config = qcf_config
        self._result_data = result_data

        self._compute: Optional[MockTestingComputeManager] = None
        self._compute_thread = None

        self._finalizer = None

    # Classmethod because finalizer can't handle bound methods
    @classmethod
    def _stop(cls, compute, compute_thread):
        if compute is not None:
            compute.stop()
            compute_thread.join()

    def start(self, manual_updates) -> None:
        if self._compute is not None:
            raise RuntimeError("Compute manager already started")
        self._compute = MockTestingComputeManager(self._qcf_config, self._result_data)
        self._compute_thread = threading.Thread(
            target=self._compute.start, kwargs={"manual_updates": manual_updates}, daemon=True
        )
        self._compute_thread.start()

        self._finalizer = weakref.finalize(
            self,
            self._stop,
            self._compute,
            self._compute_thread,
        )

    def stop(self) -> None:
        if self._finalizer is not None:
            self._finalizer()

        self._compute = None
        self._compute_thread = None

    def is_alive(self) -> bool:
        return self._compute_thread.is_alive()


def populate_db(storage_socket: SQLAlchemySocket):
    # explicitly load enough so we have to do chunking on the return
    id_0, result_data_0 = submit_opt_test_data(storage_socket, "opt_psi4_methane_sometraj", "tag0", PriorityEnum.normal)
    id_1, result_data_1 = submit_sp_test_data(storage_socket, "sp_psi4_water_gradient", "tag1", PriorityEnum.high)
    id_2, result_data_2 = submit_sp_test_data(storage_socket, "sp_psi4_water_hessian", "tag2", PriorityEnum.high)
    id_3, result_data_3 = submit_sp_test_data(storage_socket, "sp_psi4_peroxide_energy_wfn", "tag3", PriorityEnum.high)
    id_4, result_data_4 = submit_sp_test_data(storage_socket, "sp_rdkit_water_energy", "tag4", PriorityEnum.normal)
    id_5, result_data_5 = submit_sp_test_data(storage_socket, "sp_psi4_benzene_energy_2", "tag5", PriorityEnum.normal)
    id_6, result_data_6 = submit_sp_test_data(storage_socket, "sp_psi4_water_energy", "tag6", PriorityEnum.normal)
    all_id = [id_0, id_1, id_2, id_3, id_4, id_5, id_6]

    result_data = {
        id_0: result_data_0,
        id_1: result_data_1,
        id_2: result_data_2,
        id_3: result_data_3,
        id_4: result_data_4,
        id_5: result_data_5,
        id_6: result_data_6,
    }

    return all_id, result_data



from __future__ import annotations

import json
import os
import shutil
import subprocess
import time
from typing import Optional, Dict, Tuple, List

from qcfractalcompute.compress import compress_result
from .models import AppTaskResult

_apptainer_cmd = None


def get_apptainer_cmd() -> str:
    global _apptainer_cmd

    if _apptainer_cmd is not None:
        return _apptainer_cmd

    _apptainer_cmd = shutil.which("apptainer")
    if _apptainer_cmd is None:
        _apptainer_cmd = shutil.which("singularity")
    if _apptainer_cmd is None:
        raise RuntimeError("apptainer or singularity not found in PATH")

    return _apptainer_cmd


def run_apptainer(sif_path: str, command: List[str], volumes: List[Tuple[str, str]]) -> AppTaskResult:
    cmd = [get_apptainer_cmd()]

    volumes_tmp = [f"{v[0]}:{v[1]}" for v in volumes]
    cmd.extend(["run", "--bind", ",".join(volumes_tmp), sif_path])
    cmd.extend(command)

    time_0 = time.time()
    proc_result = subprocess.run(cmd, capture_output=True, text=True)
    time_1 = time.time()

    if proc_result.returncode == 0:
        ret = json.loads(proc_result.stdout)
    else:
        msg = (
            f"Running in apptainer failed with error code {proc_result.returncode}\n"
            f"stdout: {proc_result.stdout}\n"
            f"stderr: {proc_result.stderr}"
        )

        ret = {"success": False, "error": {"error_type": "RuntimeError", "error_message": msg}}

    return AppTaskResult(
        success=ret["success"],
        walltime=time_1 - time_0,
        result_compressed=compress_result(ret),
    )


def run_conda_subprocess(
    conda_env_name: Optional[str], cmd: List[str], cwd: Optional[str], env: Dict[str, str]
) -> AppTaskResult:
    if cwd:
        cwd = os.path.expandvars(cwd)

    sub_env = os.environ.copy()
    sub_env.update(env)

    if conda_env_name:
        cmd = ["conda", "run", "-n", conda_env_name] + cmd

    time_0 = time.time()
    proc_result = subprocess.run(cmd, capture_output=True, text=True, cwd=cwd, env=sub_env)
    time_1 = time.time()

    if proc_result.returncode == 0:
        ret = json.loads(proc_result.stdout)
    else:
        msg = (
            f"Subprocess failed with error code {proc_result.returncode}\n"
            f"stdout: {proc_result.stdout}\n"
            f"stderr: {proc_result.stderr}"
        )

        ret = {"success": False, "error": {"error_type": "RuntimeError", "error_message": msg}}

    return AppTaskResult(
        success=ret["success"],
        walltime=time_1 - time_0,
        result_compressed=compress_result(ret),
    )



"""
Helpers for compressing data to send back to the server
"""

from typing import Dict, Any

import numpy

from qcportal.compression import CompressionEnum, compress


def _compress_common(result: Dict[str, Any]):
    """
    Compresses outputs of an AtomicResult or OptimizationResult, storing them in extras
    """

    stdout = result.get("stdout", None)
    stderr = result.get("stderr", None)
    error = result.get("error", None)

    compressed_outputs = {}

    if stdout is not None:
        result["extras"].setdefault("_qcfractal_compressed_outputs", {})
        new_stdout, ctype, clevel = compress(stdout, CompressionEnum.zstd)
        compressed_outputs["stdout"] = {"compression_type": ctype, "compression_level": clevel, "data": new_stdout}
        result["stdout"] = None

    if stderr is not None:
        result["extras"].setdefault("_qcfractal_compressed_outputs", {})
        new_stderr, ctype, clevel = compress(stderr, CompressionEnum.zstd)
        compressed_outputs["stderr"] = {"compression_type": ctype, "compression_level": clevel, "data": new_stderr}
        result["stderr"] = None

    if error is not None:
        result["extras"].setdefault("_qcfractal_compressed_outputs", {})
        new_error, ctype, clevel = compress(error.dict(), CompressionEnum.zstd)
        compressed_outputs["error"] = {"compression_type": ctype, "compression_level": clevel, "data": new_error}
        result["error"] = None

    if compressed_outputs:
        result["extras"]["_qcfractal_compressed_outputs"] = compressed_outputs


def _compress_native_files(result: Dict[str, Any]):
    """
    Compresses outputs and native files, storing them in extras
    """

    native_files = result.get("native_files", None)
    if not native_files:
        return result

    compressed_nf = {}
    for name, data in native_files.items():
        nf, ctype, clevel = compress(data, CompressionEnum.zstd)
        compressed_nf[name] = {"compression_type": ctype, "compression_level": clevel, "data": nf}

    result["native_files"] = {}
    result["extras"]["_qcfractal_compressed_native_files"] = compressed_nf


def _compress_optimizationresult(result: Dict[str, Any]):
    """
    Compresses outputs inside an OptimizationResult, storing them in extras

    Outputs for the AtomicResults stored in the trajectory will be stored in the extras for that AtomicResult
    """

    # Handle the trajectory
    if result.get("trajectory", None):
        for x in result["trajectory"]:
            _compress_common(x)

    # Now handle the outputs of the optimization itself
    _compress_common(result)


def _convert_numpy(obj):
    if isinstance(obj, dict):
        return {k: _convert_numpy(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple, set)):
        return [_convert_numpy(v) for v in obj]
    elif isinstance(obj, numpy.ndarray):
        if obj.shape:
            return obj.ravel().tolist()
        else:
            return obj.tolist()
    else:
        return obj


def compress_result(result: Dict[str, Any]) -> bytes:
    """
    Compress outputs and native files inside results, storing them in extras. Then compress the whole result

    Outputs and native files are put into the database compressed, so no decompression is done
    until someone requests them (and then decompression happens on the client)

    The compressed outputs are stored in extras. For OptimizationResult, the outputs for the optimization
    are stored in the extras field of the OptimizationResult, while the outputs for the trajectory
    are stored in the extras field for the AtomicResults within the trajectory
    """

    result = _convert_numpy(result)
    schema_type = result.get("schema_name", None)

    if schema_type == "qcschema_output":
        _compress_common(result)
        _compress_native_files(result)
    elif schema_type == "qcschema_optimization_output":
        _compress_optimizationresult(result)
    elif schema_type == "qca_generic_task_result":
        _compress_common(result)
    else:
        pass

    # Compress the whole thing
    r, _, _ = compress(result, CompressionEnum.zstd)
    return r



"""
A command line interface to the qcfractal server.
"""

import argparse
import logging
import signal
import sys

from . import __version__
from .compute_manager import ComputeManager
from .config import read_configuration


def main():
    parser = argparse.ArgumentParser(description="A CLI for a QCFractal QueueManager")
    parser.add_argument("--version", action="version", version=f"{__version__}")
    parser.add_argument("--verbose", action="store_true", help="Print verbose output")
    parser.add_argument("--config", type=str, default=None)

    args = parser.parse_args()

    # Universal formatting for logs
    formatter = logging.Formatter("[%(asctime)s] %(levelname)8s: %(name)s: %(message)s", "%Y-%m-%d %H:%M:%S %Z")

    # Set up a log handler that is used before the logfile is set up
    log_handler = logging.StreamHandler(sys.stdout)
    log_handler.setFormatter(formatter)

    # If the user wants verbose output (particularly about startup of all the commands), then set logging level
    # to be DEBUG
    if args.verbose:
        logging.basicConfig(level="DEBUG", handlers=[log_handler], format="%(levelname)s: %(name)s: %(message)s")
    else:
        logging.basicConfig(level="INFO", handlers=[log_handler], format="%(levelname)s: %(message)s")

    manager_config = read_configuration([args.config])

    # Adjust the logging level to what was in the config
    logging.getLogger().setLevel(manager_config.loglevel)

    if manager_config.logfile:
        print("*" * 10)
        print(f"Logging to file {manager_config.logfile}, logging level {manager_config.loglevel}")
        print("*" * 10)
        log_handler = logging.FileHandler(manager_config.logfile)
        log_handler.setFormatter(formatter)
        logging.getLogger().handlers = [log_handler]

    manager = ComputeManager(manager_config)

    # Catch some signals to gracefully shutdown
    for signame in {"SIGHUP", "SIGINT", "SIGTERM"}:

        def stop(*args, **kwargs):
            logging.info(f"Received signal {signame}, shutting down")
            manager.stop()

        signal.signal(getattr(signal, signame), stop)

    # Blocks until signal
    manager.start()


if __name__ == "__main__":
    main()



from __future__ import annotations

import yaml

import pytest
from qcfractalcompute.config import SlurmExecutorConfig


@pytest.mark.parametrize("time_str", ["02:01:59", "72:00:00", "10:00:00"])
def test_manager_config_walltime(time_str):
    # Walltimes in a yaml can be parsed as seconds. Test we convert that correctly

    # Walltime as a string
    config_yaml = f"""
        queue_tags:
          - '*'
        cores_per_worker: 1
        memory_per_worker: 1.0
        max_nodes: 1
        workers_per_node: 1
        walltime: "{time_str}"
    """

    config = yaml.load(config_yaml, yaml.SafeLoader)
    executor_config = SlurmExecutorConfig(**config)
    assert executor_config.walltime == time_str

    # Walltime without quotes (gets converted by yaml to int)
    config_yaml = f"""
        queue_tags:
          - '*'
        cores_per_worker: 1
        memory_per_worker: 1.0
        max_nodes: 1
        workers_per_node: 1
        walltime: {time_str}
    """

    config = yaml.load(config_yaml, yaml.SafeLoader)
    executor_config = SlurmExecutorConfig(**config)
    assert executor_config.walltime == time_str



from __future__ import annotations

import os
import tempfile
from typing import Optional

from qcfractalcompute.config import ExecutorConfig
from .models import AppTaskResult


def geometric_nextchain_conda_app(
    record_id: int,
    function_kwargs_compressed: bytes,
    executor_config: ExecutorConfig,
    conda_env_name: Optional[str],
) -> AppTaskResult:
    import json
    from qcportal.compression import decompress, CompressionEnum
    from qcfractalcompute.apps.helpers import run_conda_subprocess
    from qcfractalcompute.run_scripts import get_script_path

    script_path = get_script_path("geometric_nextchain.py")

    env = os.environ.copy()
    env["OMP_NUM_THREADS"] = str(executor_config.cores_per_worker)
    env["MKL_NUM_THREADS"] = str(executor_config.cores_per_worker)

    function_kwargs = decompress(function_kwargs_compressed, CompressionEnum.zstd)

    with tempfile.NamedTemporaryFile("w") as f:
        json.dump(function_kwargs, f)
        f.flush()

        cmd = ["python3", script_path, str(record_id), f.name]
        return run_conda_subprocess(conda_env_name, cmd, executor_config.scratch_directory, env)


def geometric_nextchain_apptainer_app(
    record_id: int,
    function_kwargs_compressed: bytes,
    executor_config: ExecutorConfig,
    sif_path: str,
) -> AppTaskResult:
    import json
    from qcportal.compression import decompress, CompressionEnum
    from qcfractalcompute.apps.helpers import run_apptainer
    from qcfractalcompute.run_scripts import get_script_path

    script_path = get_script_path("geometric_nextchain.py")

    function_kwargs = decompress(function_kwargs_compressed, CompressionEnum.zstd)

    with tempfile.NamedTemporaryFile("w") as f:
        json.dump(function_kwargs, f)
        f.flush()

        volumes = [(script_path, "/geometric_nextchain.py"), (f.name, "/input.json")]
        cmd = ["python3", "/geometric_nextchain.py", str(record_id), "/input.json"]

        return run_apptainer(sif_path, command=cmd, volumes=volumes)



from __future__ import annotations

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic

from qcportal.compression import decompress, CompressionEnum


class AppTaskResult(pydantic.BaseModel):
    """
    A result that is returned from a parsl future
    """

    success: bool
    walltime: float
    result_compressed: bytes

    @property
    def result(self):
        return decompress(self.result_compressed, CompressionEnum.zstd)



import importlib
import io
import json
import operator
import sys
import time
from contextlib import redirect_stdout, redirect_stderr

if __name__ == "__main__":
    function_call_data_file = sys.argv[1]

    with open(function_call_data_file, "r") as f:
        function_call_data = json.load(f)

    module_name, func_name = function_call_data["function"].split(".", 1)
    module = importlib.import_module(module_name)
    func = operator.attrgetter(func_name)(module)

    with redirect_stdout(io.StringIO()) as rdout:
        with redirect_stderr(io.StringIO()) as rderr:
            start_time = time.time()
            try:
                results = func(**function_call_data["function_kwargs"])
            except Exception as e:
                ret = {
                    "error": {
                        "error_type": type(e).__name__,
                        "error_message": str(e),
                    },
                    "success": False,
                }

            end_time = time.time()

    stdout = rdout.getvalue()
    stderr = rderr.getvalue()

    ret = {
        "id": function_call_data["record_id"],
        "success": True,
        "stdout": stdout if stdout else None,  # convert empty string to None
        "stderr": stderr if stderr else None,
        "results": results,
        "provenance": {
            "creator": function_call_data["function"],
            "walltime": (end_time - start_time),
        },
    }

    print(json.dumps(ret))



from __future__ import annotations

import os
import tempfile
from typing import Optional

from qcfractalcompute.config import ExecutorConfig
from .models import AppTaskResult


def qcengine_conda_app(
    record_id: int,
    function_kwargs_compressed: bytes,
    executor_config: ExecutorConfig,
    conda_env_name: Optional[str],
) -> AppTaskResult:
    import json
    from qcportal.compression import decompress, CompressionEnum
    from qcfractalcompute.apps.helpers import run_conda_subprocess
    from qcfractalcompute.run_scripts import get_script_path

    script_path = get_script_path("qcengine_compute.py")

    # This function handles both compute and compute_procedure
    # Record id can be ignored, but is here for consistency with other apps

    if executor_config.scratch_directory:
        scratch_directory = os.path.expandvars(executor_config.scratch_directory)
        scratch_directory = os.path.expanduser(scratch_directory)
    else:
        scratch_directory = None

    qcengine_options = {}
    qcengine_options["memory"] = executor_config.memory_per_worker
    qcengine_options["ncores"] = executor_config.cores_per_worker
    qcengine_options["scratch_directory"] = scratch_directory

    function_kwargs = decompress(function_kwargs_compressed, CompressionEnum.zstd)
    function_kwargs = {**function_kwargs, "task_config": qcengine_options}

    with tempfile.NamedTemporaryFile("w") as f:
        json.dump(function_kwargs, f)
        f.flush()

        cmd = ["python3", script_path, f.name]
        return run_conda_subprocess(conda_env_name, cmd, executor_config.scratch_directory, {})


def qcengine_apptainer_app(
    record_id: int,
    function_kwargs_compressed: bytes,
    executor_config: ExecutorConfig,
    sif_path: str,
) -> AppTaskResult:
    import json
    from qcportal.compression import decompress, CompressionEnum
    from qcfractalcompute.apps.helpers import run_apptainer
    from qcfractalcompute.run_scripts import get_script_path

    script_path = get_script_path("qcengine_compute.py")

    # This function handles both compute and compute_procedure
    # Record id can be ignored, but is here for consistency with other apps

    if executor_config.scratch_directory:
        scratch_directory = os.path.expandvars(executor_config.scratch_directory)
        scratch_directory = os.path.expanduser(scratch_directory)
    else:
        scratch_directory = None

    qcengine_options = {}
    qcengine_options["memory"] = executor_config.memory_per_worker
    qcengine_options["ncores"] = executor_config.cores_per_worker
    qcengine_options["scratch_directory"] = scratch_directory

    function_kwargs = decompress(function_kwargs_compressed, CompressionEnum.zstd)
    function_kwargs = {**function_kwargs, "task_config": qcengine_options}

    with tempfile.NamedTemporaryFile("w") as f:
        json.dump(function_kwargs, f)
        f.flush()

        volumes = [(script_path, "/qcengine_compute.py"), (f.name, "/input.json")]
        cmd = ["python3", "/qcengine_compute.py", "/input.json"]

        return run_apptainer(sif_path, command=cmd, volumes=volumes)



import json
import sys

import qcengine

# From psi4/driver/p4util/python_helpers.py
# https://github.com/psi4/psi4/blob/master/psi4/driver/p4util/python_helpers.py
_qcvar_transitions = {
    "scsn-mp2 correlation energy": "scs(n)-mp2 correlation energy",
    "scsn-mp2 total energy": "scs(n)-mp2 total energy",
    "mayer_indices": "mayer indices",
    "wiberg_lowdin_indices": "wiberg lowdin indices",
    "lowdin_charges": "lowdin charges",
    "mulliken_charges": "mulliken charges",
    "(at) correction energy": "a-(t) correction energy",
    "ccsd(at) total energy": "a-ccsd(t) total energy",
    "ccsd(at) correlation energy": "a-ccsd(t) correlation energy",
    "cp-corrected 2-body interaction energy": "cp-corrected interaction energy through 2-body",
    "cp-corrected 3-body interaction energy": "cp-corrected interaction energy through 3-body",
    "cp-corrected 4-body interaction energy": "cp-corrected interaction energy through 4-body",
    "cp-corrected 5-body interaction energy": "cp-corrected interaction energy through 5-body",
    "nocp-corrected 2-body interaction energy": "nocp-corrected interaction energy through 2-body",
    "nocp-corrected 3-body interaction energy": "nocp-corrected interaction energy through 3-body",
    "nocp-corrected 4-body interaction energy": "nocp-corrected interaction energy through 4-body",
    "nocp-corrected 5-body interaction energy": "nocp-corrected interaction energy through 5-body",
    "vmfc-corrected 2-body interaction energy": "vmfc-corrected interaction energy through 2-body",
    "vmfc-corrected 3-body interaction energy": "vmfc-corrected interaction energy through 3-body",
    "vmfc-corrected 4-body interaction energy": "vmfc-corrected interaction energy through 4-body",
    "vmfc-corrected 5-body interaction energy": "vmfc-corrected interaction energy through 5-body",
    "counterpoise corrected total energy": "cp-corrected total energy",
    "counterpoise corrected interaction energy": "cp-corrected interaction energy",
    "non-counterpoise corrected total energy": "nocp-corrected total energy",
    "non-counterpoise corrected interaction energy": "nocp-corrected interaction energy",
    "valiron-mayer function couterpoise total energy": "valiron-mayer function counterpoise total energy",  # note misspelling
    "valiron-mayer function couterpoise interaction energy": "vmfc-corrected interaction energy",  # note misspelling
}

if __name__ == "__main__":
    function_kwargs_file = sys.argv[1]

    with open(function_kwargs_file, "r") as f:
        function_kwargs = json.load(f)

    if "procedure" in function_kwargs:
        ret = qcengine.compute_procedure(**function_kwargs)
    else:
        ret = qcengine.compute(**function_kwargs)

        # Hacky - handle keys in qcvars
        if ret.success and ret.extras and "qcvars" in ret.extras:
            # Make qcvars keys all lowercase
            ret.extras["qcvars"] = {k.lower(): v for k, v in ret.extras["qcvars"].items()}

            # Remove any from qcvars that are in properties (but with underscores)
            prop_dict = ret.properties.dict()
            to_delete = [k for k in ret.extras["qcvars"].keys() if k.replace(" ", "_") in prop_dict.keys()]
            for k in to_delete:
                ret.extras["qcvars"].pop(k)

            # Replace any names with underscores (and other modifications)
            ret.extras["qcvars"] = {_qcvar_transitions.get(k, k): v for k, v in ret.extras["qcvars"].items()}

    print(json.dumps(ret.dict(encoding="json")))



"""
Compute worker and manager for QCArchive/QCFractal
"""

from importlib.metadata import version

__version__ = version("qcfractalcompute")

from .compute_manager import ComputeManager
from .generic_wrapper import wrap_generic_function
